open Printf;;
open Message;;

type t = {
  lexbuf : Lexing.lexbuf;
  mutable name : string;
}

type env = {
  env_wsk    : Message.world_skeleton;
  env_topo   : Topology.topo;
  mutable env_info   : Info.info;
  env_drawer : Draw.t;
}

(*
let ensure_room_array a n = 
  if (n < Array.length a) then a
  else
    let b = Array.create (2 * n + 10) None in
    Array.blit a 0 b 0 (Array.length a);
    b

let ensure_room_old_infos env n =
  env.env_old_infos <- ensure_room_array env.env_old_infos n

let register_old_info env info =
  let wnum = Info.current_world env.env_info in
  ensure_room_old_infos env wnum;
  env.env_old_infos.(wnum) <- Some info

let get_old_info env wnum =
  if (wnum < 0) || (wnum >= Array.length env.env_old_infos) 
  then None
  else env.env_old_infos.(wnum)
*)

type mov = (int * (Message.ptype * Message.nod)) list

let string_of_mov (ptype, nod) = 
  Message.ptype_to_string ptype ^ " " ^ nod.nod_loc

let dump_moves moves =
  List.iter (fun (i,move) ->
	       Printf.eprintf "dump_mov for %i: %s\n" i (string_of_mov move)) moves;
  flush stderr

type 'a cops_fun = {
  cops_init : t -> env -> 'a;
  cops_clean_job  : t -> env -> world -> 'a -> (offer * mov * int list * 'a);
  cops_dirty_job  : t -> env -> world -> 'a -> (mov * 'a);
  cops_dirty_r_job : t -> env -> world -> 'a -> 'a;
  cops_ptype : Message.ptype;
}

let time_mark me s = 
(*  Printf.eprintf "[%f] %s (%s)\n" (Unix.gettimeofday ()) s
    me.name; flush stderr *)
  ()

let timer = ref 0.

let timer_start () =
(*  timer := Unix.gettimeofday (); *)
  ()

let timer_stop s =
(*  let t = Unix.gettimeofday () in
  Printf.eprintf "Timer %s: %05i ms\n" s 
    (int_of_float (1000. *. (t -. !timer))); *)
  ()


let cop_register t name typ =
  Message.register {
    Message.reg_bot = name;
    Message.reg_ptype = typ;
  };
  let m = Parser.world_skeleton Lexer.token t.lexbuf in
  time_mark t "world skeleton received";
  timer_start ();
  m

let get_next t = Parser.world_or_game_over Lexer.token t.lexbuf

let cop_first t =
  let m = get_next t in
  time_mark t "first world received";
  timer_start ();
  m

let cop_inform t env inf =
  Message.inform inf;
  time_mark t "inform sent";
  timer_stop "inform";
  let informs = Parser.from_inform Lexer.token t.lexbuf in
  time_mark t "informs received";
  timer_start ();
  Info.add_informs env.env_info informs;
  informs

let cop_r_inform t env inf = 
  Message.inform inf;
  Parser.from_inform Lexer.token t.lexbuf


let cop_plan t plan =
  Message.plan plan;
  time_mark t "plan sent";
  timer_stop "plan";
  let m = Parser.from_plan Lexer.token t.lexbuf in
  time_mark t "plans received";
  timer_start ();
  m

let cop_vote t v =
  Message.vote v;
  time_mark t "vote sent";
  timer_stop "vote";
  let m = Parser.vote_res Lexer.token t.lexbuf in
  time_mark t "votes received";
  timer_start ();
  m


let clean_cop_move t env w (offer,movs,accuse) =
  let moves = Topology.can_i_move env.env_topo w movs in
  let accuse = List.map (fun i -> env.env_topo.Topology.cops_name.(i)) accuse in
  Message.clean_cop_move 
    { Message.mov_offer = offer;
      Message.mov_accuse = accuse;
      Message.mov_moves = moves };
  time_mark t "move sent";
  timer_stop "move";
  let m = get_next t in
  time_mark t "next world received";
  timer_start ();
  m

let dirty_cop_move t env w movs =
  let moves = Topology.can_i_move env.env_topo w movs in
  Message.dirty_cop_move moves;
  time_mark t "move sent";
  timer_stop "move";
  let m = get_next t in
  time_mark t "next world received";
  timer_start ();
  match m with
    | Message.Game_over -> exit 0
    | Message.World w -> w


let cops = Hashtbl.create 16

let register_cop n f =
  assert(not (Hashtbl.mem cops n));
  Hashtbl.add cops n ((Obj.magic f) : 'a cops_fun)

let create_env = fun me name ptype ->
  let wsk = cop_register me name ptype in
  let topo = Topology.read_skel wsk in
  me.name <- topo.Topology.name;
  let drawer = Draw.init topo in
  let info = Info.init topo drawer in
    { env_wsk = wsk;
      env_topo = topo;
      env_info = info;
      env_drawer = drawer;
    }

let run_cop ic name =
  let t = { lexbuf = Lexing.from_channel ic; name = "anonymous" } in
  let f =
    try Hashtbl.find cops name
    with Not_found -> failwith (Printf.sprintf "Cops `%s' not implemented" name)
  in
  let env = create_env t name f.cops_ptype in

  let rec message user_infos = function
    | Message.Game_over -> ()
    | Message.World w -> world user_infos w
  and world user_infos w =
    env.env_info <- Info.add_world env.env_info w;
    if Info.am_i_dirty env.env_info then
      let movs,user_infos = f.cops_dirty_job t env w user_infos in
      let w' = dirty_cop_move t env w movs in
      (* Ici: intgrer le monde w' ??? *)
      let user_infos = f.cops_dirty_r_job t env w' user_infos in
      message user_infos (get_next t)
    else
      let (offer,movs,acc, user_infos) = f.cops_clean_job t env w user_infos in
      message user_infos (clean_cop_move t env w (offer,movs,acc))
  in
  let user_infos = f.cops_init t env in
    message user_infos (cop_first t)
