open Topology
open Message

type t = {
  topo: topo;
  size_x: int;
  size_y: int;

  max_x: int;
  max_y: int;
  n: int;
}

let enabled = ref true;;

let open_graph s = if !enabled then Graphics.open_graph s;;
let set_window_title s = if !enabled then Graphics.set_window_title s;;
let size_x () = if !enabled then Graphics.size_x () else 100;;
let size_y () = if !enabled then Graphics.size_y () else 100;;
let set_color c = if !enabled then Graphics.set_color c;;
let draw_circle x y r = if !enabled then Graphics.draw_circle x y r;;
let fill_circle x y r = if !enabled then Graphics.fill_circle x y r;;
let black = Graphics.black;;
let blue = Graphics.blue;;
let yellow = Graphics.yellow;;
let green = Graphics.green;;
let red = Graphics.red;;
let moveto x y = if !enabled then Graphics.moveto x y;;
let lineto x y = if !enabled then Graphics.lineto x y;;
let curveto a b c = if !enabled then Graphics.curveto a b c;;
let clear_graph () = if !enabled then Graphics.clear_graph ();;
let read_key () = if !enabled then Graphics.read_key () else 'q';;
let close_graph () = if !enabled then Graphics.close_graph ();;
let draw_string s = if !enabled then Graphics.draw_string s;;
let key_pressed () = if !enabled then Graphics.key_pressed () else false;;
let auto_synchronize b = if !enabled then Graphics.auto_synchronize b else ()
let synchronize () = if !enabled then Graphics.synchronize () else ()

module Graphics = struct end

let init ?(sx=800) ?(sy=600) topo =
  open_graph (Printf.sprintf ":0 %ix%i" sx sy);
  set_window_title topo.name;
  let max_x = ref 0 and max_y = ref 0 in
  Array.iter (fun node ->
		max_x := max !max_x node.nod_x;
		max_y := max !max_y node.nod_y;
	     ) topo.t_nods_alpha;
  let max_x = (10 + !max_x) * 11 / 10 in
  let max_y = (10 + !max_y) * 11 / 10 in
  { n = Array.length topo.t_nods_alpha;
    topo = topo; 
    size_x = size_x (); size_y = size_y ();
    max_x = max_x; max_y = max_y }

let coord p n =
  let x = n.nod_x and y = n.nod_y in
  (10+x) * p.size_x / p.max_x,
  p.size_y - (10+y) * p.size_y / p.max_y

let coord_i p i =
  coord p p.topo.t_nods.(i)

let circle c x y r =
  set_color c; draw_circle x y r
let fcircle c x y r =
  set_color c; fill_circle x y r;
  circle black x y r

let draw_nodei p i =
  let x,y = coord_i p i in
  match p.topo.t_nods.(i).nod_tag with
    | Hq -> fcircle blue x y 5
    | Bank -> fcircle yellow x y 5
    | _ -> circle black x y 2

let draw_edge p col i j =
  let x1,y1 = coord_i p i and x2,y2 = coord_i p j in
  set_color col;
  moveto x1 y1;
  lineto x2 y2

let has_foot_edge p i j =
  List.mem j p.topo.t_neigh_f.(i)
let has_car_edge p i j =
  List.mem j p.topo.t_neigh_c.(i)

let draw_oneway_edge p col i j =
  let x1,y1 = coord_i p i and x2,y2 = coord_i p j in
  set_color col;
  moveto x1 y1;
  if not (has_car_edge p i j) then
    curveto (x1-(y1-y2)/10,y1-(x2-x1)/10) (x2-(y1-y2)/10,y2-(x2-x1)/10) (x2,y2)

let draw_title p s =
  moveto 10 (p.size_y - 20);
  draw_string s

let draw_nodes p =
  for i = 0 to p.n - 1 do draw_nodei p i done

let draw_map p m color =
  Array.iteri (fun i l -> List.iter (draw_edge p color i) l) m

let draw_foot_map p color =
  draw_map p p.topo.t_neigh_f color

let draw_oneway_map p color =
  Array.iteri (fun i l -> List.iter (draw_oneway_edge p color i) l) p.topo.t_neigh_f

let display_node_sets p ?(title="") ?(wait=false) ns1 ns2 =
  let f1 n =
    let x,y = coord p n in
    fcircle green x y 8 in
  let f2 n =
    let x,y = coord p n in
    fcircle red x y 5 in

  auto_synchronize false;
  clear_graph ();
  draw_title p title;
  draw_foot_map p blue;
  draw_nodes p;
  NS.iter f1 ns1;
  NS.iter f2 ns2;
  synchronize ();
  auto_synchronize true;
  if wait then ignore (read_key ());
;;

let wait_q _ =  
  while read_key () <> 'q' do () done;
  close_graph ()


let draw_label p n s =
  let x,y = coord p n in
  moveto (x + 4) (y + 4);
  draw_string s

let pause p = 
  ignore (read_key ())

let may_pause p =
  if key_pressed () then (ignore (read_key ()); ignore (read_key ()))
