(* $Id: topology.ml,v 1.32.4.4 2005/07/10 10:11:26 frisch Exp $ *)

open Message;;
open Printf;;

module Onodes = struct
  type t = nod;;
  let compare n1 n2 = compare n1.nod_num n2.nod_num;;
end;;

module NS = Set.Make (Onodes);;

type cop = int

type topo = {
  t_nods : nod array;
  t_nods_alpha : nod array;
  t_neigh_f : int list array;
  t_neigh_c : int list array;
  t_neigh_c_inv : int list array;

  t_dist_f : int array Lazy.t array;
  t_dist_c : int array Lazy.t array;

  t_dist_f_1 : NS.t array;
  t_dist_f_2 : NS.t array;
  t_dist_c_1 : NS.t array;

  t_banks : NS.t;
  all_nodes : NS.t;

  ncops : int;
  cops_name : string array;
  cops_number : (string * int) list;

  initial_robber : nod;
  robber_name : string;

  hq : nod;

  name : string;
}

let nodes_count = fun topo -> Array.length topo.t_nods

let cop_name topo i = topo.cops_name.(i)
let cop_number topo s = List.assoc s topo.cops_number

let find ar name =
  let rec dicho lo hi =
    assert (lo < hi);
    if lo+1 = hi then begin
      assert (name = ar.(lo).nod_loc);
      ar.(lo).nod_num
    end else begin
      let mid = (lo + hi) / 2 in
      if name < ar.(mid).nod_loc then dicho lo mid else dicho mid hi
    end
  in
  dicho 0 (Array.length ar)
;;

let get_node t s =
 t.t_nods.(find t.t_nods_alpha s)

let rec uniq l =
  match l with
  | [] | [_] -> l
  | x::((y::_) as t) -> if x = y then uniq t else x :: uniq t
;;

let remove_duplicates g =
  Array.map (fun l -> uniq (List.sort compare l)) g
;;

let read_skel s =
  let size = List.length s.wsk_nods in
  let cmp n1 n2 = compare n1.nod_loc n2.nod_loc in
  let nods_alpha_list = List.sort cmp s.wsk_nods in
  let nods_alpha = Array.of_list nods_alpha_list in
  let cmp n1 n2 = compare n1.nod_num n2.nod_num in
  let nods_list = List.sort cmp s.wsk_nods in
  let nods = Array.of_list nods_list in
  let neigh_f = Array.make size [] in
  let neigh_c = Array.make size [] in
  let neigh_c_inv = Array.make size [] in
  let add_edge g src dst =
    let srci = find nods_alpha src in
    let dsti = find nods_alpha dst in
    g.(srci) <- dsti :: g.(srci);
  in
  let do_edge e =
    match e.edg_type with
    | Car -> add_edge neigh_c e.edg_src e.edg_dst;
             add_edge neigh_c_inv e.edg_dst e.edg_src;
    | Foot -> add_edge neigh_c e.edg_src e.edg_dst;
              add_edge neigh_c_inv e.edg_dst e.edg_src;
              add_edge neigh_f e.edg_src e.edg_dst;
              add_edge neigh_f e.edg_dst e.edg_src;
  in
  List.iter do_edge s.wsk_edgs;
  let neigh_f = remove_duplicates neigh_f
  and neigh_c = remove_duplicates neigh_c in

  let dist_f_1 =
    Array.mapi
      (fun i _ ->
	 List.fold_left
	   (fun accu i -> NS.add nods.(i) accu) NS.empty neigh_f.(i)
      ) nods in
  let dist_c_1 =
    Array.mapi
      (fun i _ ->
	 List.fold_left
	   (fun accu i -> NS.add nods.(i) accu) NS.empty neigh_c.(i)
      ) nods in
  let dist_f_2 =
    Array.mapi
      (fun i d1 ->
	 let e =
	   List.fold_left
	     (fun accu n -> NS.union accu (dist_f_1.(n))) NS.empty neigh_f.(i) in
	 NS.remove nods.(i) (NS.diff e d1)
      ) dist_f_1
  in

  let all_nodes = Array.fold_right NS.add nods NS.empty in

  let cops = Array.of_list s.wsk_cops in
  { t_nods = nods;
    t_nods_alpha = nods_alpha;
    t_neigh_f = neigh_f;
    t_neigh_c = neigh_c;
    t_neigh_c_inv = remove_duplicates neigh_c_inv;

    t_dist_f = Distance.distances neigh_f;
    t_dist_c = Distance.distances neigh_c;

    t_dist_f_1 = dist_f_1;
    t_dist_f_2 = dist_f_2;
    t_dist_c_1 = dist_c_1;

    t_banks = NS.filter (fun n -> n.nod_tag = Bank) all_nodes;

    all_nodes = all_nodes;
    ncops = List.length s.wsk_cops;
    cops_name = cops;
    cops_number = Array.to_list (Array.mapi (fun i s -> (s,i)) cops);

    initial_robber = List.find (fun n -> n.nod_tag = Robber_start) nods_list;
    robber_name = s.wsk_robber;
    name = s.wsk_name;
    
    hq = List.find (fun n -> n.nod_tag = Hq) nods_list;
  }
;;

let dump_neighbours adj =
  for i = 0 to Array.length adj - 1 do
    eprintf "  %d ->" i;
    List.iter (fun x -> eprintf " %d" x) adj.(i);
    eprintf "\n";
  done;
;;

let dump t =
  let size = Array.length t.t_nods in
  assert (size = Array.length t.t_nods_alpha);
  assert (size = Array.length t.t_neigh_f);
  assert (size = Array.length t.t_neigh_c);
  assert (size = Array.length t.t_neigh_c_inv);
  eprintf "nodes:\n";
  for i = 0 to size - 1 do
    eprintf "  %d %s %s\n" i t.t_nods.(i).nod_loc
            (tag_to_string t.t_nods.(i).nod_tag);
  done;
  eprintf "neighbours by foot:\n";
  dump_neighbours t.t_neigh_f;
  eprintf "neighbours by car:\n";
  dump_neighbours t.t_neigh_c;
  eprintf "reverse neighbours by car:\n";
  dump_neighbours t.t_neigh_c_inv;
  flush stderr;
;;


let dist_f topo i j = (Lazy.force topo.t_dist_f.(i.nod_num)).(j.nod_num)
let dist_c topo i j = (Lazy.force topo.t_dist_c.(i.nod_num)).(j.nod_num)


let dist_f_1 topo n = topo.t_dist_f_1.(n.nod_num)
let dist_c_1 topo n = topo.t_dist_c_1.(n.nod_num)
let dist_f_2 topo n = topo.t_dist_f_2.(n.nod_num)

let extend topo s =
  List.fold_left (fun accu n -> NS.union accu (dist_f_1 topo n)) s (NS.elements s)

let extend_n topo n s =
  match n with
    | 0 -> s
    | 1 -> extend topo s 
    | _ ->
	let accu = ref NS.empty in
	Distance.leq_n topo.t_neigh_f n 
	  (List.map (fun n -> n.nod_num) (NS.elements s) )
	  (fun x -> accu := NS.add topo.t_nods.(x) !accu);
	!accu

let was_at topo n i1 i2 =
  let d = Lazy.force topo.t_dist_f.(n.nod_num) in
  let delta = (i2 + 1 - i1) / 2 in
  NS.filter (fun n' -> d.(n'.nod_num) <= delta) topo.all_nodes

let smell topo n t sm =
  match t,sm with
    | Cop_car,1 -> topo.t_dist_c_1.(n.nod_num)
    | Cop_foot,1 -> topo.t_dist_f_1.(n.nod_num)
    | Cop_foot,2 -> topo.t_dist_f_2.(n.nod_num)
    | _ -> assert false

let player_smell topo cops sm = match cops with
  | [(n,t)] -> smell topo n t sm (* opt *)
  | cops ->
      let aux i = 
	List.fold_left
	  (fun accu (n,t) -> NS.union accu (smell topo n t i))
	  NS.empty cops in
      let s = match sm with
	| 1 -> aux 1
	| 2 -> NS.diff (aux 2) (aux 1)
	| _ -> assert false in
      List.fold_left (fun accu (n,t) -> NS.remove n accu) s cops


let nosmell topo n = function
  | Cop_car -> NS.add n topo.t_dist_c_1.(n.nod_num)
  | Cop_foot -> NS.add n (NS.union topo.t_dist_f_1.(n.nod_num) topo.t_dist_f_2.(n.nod_num))
  | _ -> assert false

let player_nosmell topo cops =
  List.fold_left (fun accu (n,t) -> NS.union accu (nosmell topo n t)) NS.empty cops


let possible_dest topo n = function
  | Cop_car -> NS.add n topo.t_dist_c_1.(n.nod_num)
  | Cop_foot | Robber -> NS.add n topo.t_dist_f_1.(n.nod_num)

(*
let possible_dest_robot topo pl =
  possible_dest topo (get_node topo pl.pl_loc) pl.pl_ptype
*)

let dist_to_set topo pt set =
  let f p2 accu =
    let d = dist_f topo pt p2 in
    min d accu
  in
  NS.fold f set 10000

let dist_to_set_typ topo pt set typ =
  let f p2 accu =
    let d = match typ with
            | Cop_foot | Robber -> dist_f topo pt p2
            | Cop_car -> dist_c topo pt p2
    in
    min d accu
  in
  NS.fold f set 10000

let is_adjacent topo ptype n1 n2 =
  match ptype with
    | Cop_foot | Robber -> List.mem n1.nod_num topo.t_neigh_f.(n2.nod_num)
    | Cop_car           -> List.mem n1.nod_num topo.t_neigh_c.(n2.nod_num)

let node_of_player topo player =
  get_node topo player.pl_loc

exception BadMove

let cop_check_move topo cur_bot mode dest = 
  let cur_type = cur_bot.pl_ptype
  and cur_pos = cur_bot.pl_loc in
  let cur_node = get_node topo cur_pos in
  (* Changement de vehicule ok ? *)
  if (cur_type <> mode) && not (cur_pos = topo.hq.nod_loc) then raise BadMove;
  (* Deplacement ok selon le nouveau mode *)
  if (cur_node.nod_num <> dest.nod_num) then
    (match mode with
       | Cop_foot -> 
	   if not (List.mem dest.nod_num topo.t_neigh_f.(cur_node.nod_num))
	   then raise BadMove
       | Cop_car -> 
	   if not (List.mem dest.nod_num topo.t_neigh_c.(cur_node.nod_num))
	   then raise BadMove
       | Robber -> raise BadMove)

let cop_can_i_move topo cur_bot mode dest =
  try 
    cop_check_move topo cur_bot mode dest;
    mode,dest.nod_loc
  with BadMove ->
    let cur_type = cur_bot.pl_ptype
    and cur_pos = cur_bot.pl_loc in
    let cur_node = get_node topo cur_pos in
    Printf.eprintf "BAD MOVE !!!!!!!!!!!!!! STAY WHERE I AM !!!!!!!!!!!!!!\n";
    Printf.eprintf "(cur_loc:%s,cur_type:%s) (dest_loc:%s,dest_type:%s)\n"
      cur_pos (Message.ptype_to_string cur_type)
      dest.nod_loc (Message.ptype_to_string mode);
    cur_type,cur_pos

let is_legal topo world movs = 
  (* La liste des policiers que je controle *)
  let cops = List.filter (fun sc -> sc.sc_brain = topo.name) world.wor_sc in
  let cops = topo.name :: (List.map (fun sc -> sc.sc_cop) cops) in

  if List.length cops <> List.length  movs then false
  else
  try
    List.iter
      (fun name -> 
	 let bot =
	   List.find (fun pl -> pl.pl_bot = name) world.wor_pls in
	 let num = cop_number topo name in
	 let dest_typ,dest_loc = List.assoc num movs in
	 cop_check_move topo bot dest_typ dest_loc
      ) cops;
    true
  with Not_found | BadMove -> false

let can_i_move topo world movs =
  (* La liste des policiers que je controle *)
  let cops = List.filter (fun sc -> sc.sc_brain = topo.name) world.wor_sc in
  let cops = topo.name :: (List.map (fun sc -> sc.sc_cop) cops) in

  List.map 
    (fun name -> 
       let bot =
	 List.find (fun pl -> pl.pl_bot = name) world.wor_pls in
       let num = cop_number topo name in
       let ptype,loc = 
	 try 
	   let dest_typ,dest_loc = List.assoc num movs in
	   cop_can_i_move topo bot dest_typ dest_loc 
	 with Not_found ->
	   bot.pl_ptype,bot.pl_loc in
       { mov_loc = loc;
	 mov_ptype = ptype;
	 mov_bot = name }
    ) cops

let dump_topology world_skeleton =
  let top = read_skel world_skeleton in
    dump top
;;


